/*-
 * Copyright (C)2013 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2013 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: json2pg.c,v 1.3 2013/04/15 15:43:39 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <json/json.h>
#include <sysexits.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <getCGIparm.h>
#include <mife.h>

struct json2pg_chain;
typedef struct {
    struct json2pg_chain *next;
    char                 *n;
} json2pg_chain;

typedef struct {
    BLIN_flag  flags;
#       define JSON2PG_DUMP    0x000100
#       define JSON2PG_ODMP /* 0x000200 */ (JSON2PG_DUMP << 1)
#       define JSON2PG_FDMP /* 0x000400 */ (JSON2PG_ODMP << 1)
#       define JSON2PG_COPY    0x001000
#       define JSON2PG_COPZ /* 0x002000 */ (JSON2PG_COPY << 1)
} json2pg_cf;

static char *jtnm[] =
{ "null   "
, "boolean"
, "double "
, "int    "
, "object "
, "array  "
, "string "
, "unknown"
};

int
json2pg_dump( json2pg_cf *options
            , struct json_object *jobj
            , char *nm    /*     */
            , u_char *nnm /*                     */
            , int ex      /*        */
            , int exx     /*                   */
            ) {
#   define IFNMM (nnmm ? nnmm : "\\N")
#   define IFNM (nnmm ? nnmm : "")
    ifBLIN_FLAG = options->flags;
    u_char *pnm, *jval, *jv;
    struct json_object *val;
    struct lh_entry *entry;
    char *key, *nnmm;
    json_type jt;
    int i, e;

    ifBLIN_QX2( "+%s.%s %u<-%u\n", nm ? nm : "()", nnm ? nnm : (u_char*)"()", ex, exx);
    pnm = nnm;
    if  (!nnm) {
        nnm = (u_char*)"\\N";
        nnmm = NULL;
    } else if ((e = getCGIparmpass3(options->flags & BLIN_VERMASK, (u_char*)"copy", &nnm))) {
        ifBLIN_QW0("getCGIparmpass3 nm");
        ex = (ex > 0) ? -e : e;
        goto out;
    } else {
        if  (nm) {
            asprintf(&nnmm, "%s.%s", nm, nnm);
        } else {
            asprintf(&nnmm, "%s", nnm);
        }
        if  (!nnmm) {
            ifBLIN_QX0("no mem");
            errno = ENOMEM;
            ex = -EX_OSERR;
            goto out;
    }   }
    jt = json_object_get_type(jobj);
    switch (jt) {
    case json_type_null:
        if  (options->flags & JSON2PG_COPZ) {
            printf("%u	%u	%s	\\N\n", exx, jt, IFNMM);
        } else if (options->flags & JSON2PG_COPY) {
            printf("%u	%u	%u	%s	\\N\n", exx, ex, jt, nnm);
        } else {
            printf("%s %s\n", jtnm[jt], IFNM);
        }
        ex = exx;
        break;
    case json_type_boolean:
        if  (options->flags & JSON2PG_COPZ) {
            printf( "%u	%u	%s	%c\n"
                  , exx
                  , jt
                  , IFNMM
                  , json_object_get_boolean(jobj) ? 't' : 'f'
                  );
        } else if (options->flags & JSON2PG_COPY) {
            printf( "%u	%u	%u	%s	%c\n"
                  , exx
                  , ex
                  , jt
                  , nnm
                  , json_object_get_boolean(jobj) ? 't' : 'f'
                  );
        } else {
            printf("%s %s=%d\n", jtnm[jt], IFNM, json_object_get_boolean(jobj));
        }
        ex = exx;
        break;
    case json_type_double:
        if  (options->flags & JSON2PG_COPZ) {
            printf("%u	%u	%s	%g\n", exx, jt, IFNMM, json_object_get_double(jobj));
        } else if (options->flags & JSON2PG_COPY) {
            printf("%u	%u	%u	%s	%g\n", exx, ex, jt, nnm, json_object_get_double(jobj));
        } else {
            printf("%s %s=%g\n", jtnm[jt], IFNM, json_object_get_double(jobj));
        }
        ex = exx;
        break;
    case json_type_int:
        if  (options->flags & JSON2PG_COPZ) {
            printf("%u	%u	%s	%d\n", exx, jt, IFNMM, json_object_get_int(jobj));
        } else if (options->flags & JSON2PG_COPY) {
            printf("%u	%u	%u	%s	%d\n", exx, ex, jt, nnm, json_object_get_int(jobj));
        } else {
            printf("%s %s=%d\n", jtnm[jt], IFNM, json_object_get_int(jobj));
        }
        ex = exx;
        break;
    case json_type_object:
        if  (!(options->flags & JSON2PG_DUMP)) {
            jval = jv = (u_char*)"\\N";
        } else {
            jval = jv = (u_char*)json_object_get_string(jobj);
            if  ((e = getCGIparmpass3(options->flags & BLIN_VERMASK, (u_char*)"copy", &jv))) {
                ifBLIN_QW0("getCGIparmpass3");
                ex = (ex > 0) ? -e : e;
                goto out;
        }   }
        if  (options->flags & JSON2PG_COPZ) {
            printf("%u	%u	%s	%s\n", exx, jt, IFNMM, jv);
        } else if (options->flags & JSON2PG_COPY) {
            printf("%u	%u	%u	%s	%s\n", exx, ex, jt, nnm, jv);
        } else if (options->flags & JSON2PG_FDMP) {
            printf("%s %s=%s~\n", jtnm[jt], IFNM, json_object_get_string(jobj));
        } else if (options->flags & JSON2PG_ODMP) {
            printf("%s %s\n", jtnm[jt], IFNM);
        }
        if  (jval != jv) free(jv);
        ex = exx;
        for (entry = json_object_get_object(jobj)->head; (ex >= 0) && entry; entry = entry->next) {
            key = (char*)entry->k;
            val = (struct json_object*)entry->v;
            exx = json2pg_dump(options, val, nnmm, (u_char*)key, ex, ++exx);
            if  (nnmm) free(nnmm);
        }
        ex = exx;
        break;
    case json_type_array:
        if  (!(options->flags & JSON2PG_DUMP)) {
            jval = jv = (u_char*)"\\N";
        } else {
            jval = jv = (u_char*)json_object_get_string(jobj);
            if  ((e = getCGIparmpass3(options->flags & BLIN_VERMASK, (u_char*)"copy", &jv))) {
                ifBLIN_QW0("getCGIparmpass3");
                ex = (ex > 0) ? -e : e;
                goto out;
        }   }
        if  (options->flags & JSON2PG_COPZ) {
            printf("%u	%u	%s	%s\n", exx, jt, IFNMM, jv);
        } else if (options->flags & JSON2PG_COPY) {
            printf("%u	%u	%u	%s	%s\n", exx, ex, jt, nnm, jv);
        } else if (options->flags & JSON2PG_FDMP) {
            printf("%s %s=%s~\n", jtnm[jt], IFNM, json_object_get_string(jobj));
        } else if (options->flags & JSON2PG_ODMP) {
            printf("%s %s\n", jtnm[jt], IFNM);
        }
        if  (jval != jv) free(jv);
        ex = exx;
        for (i = 0; (ex >= 0) && (i < json_object_array_length(jobj)); ++i) {
            char *k;

            val = json_object_array_get_idx(jobj, i);
            asprintf(&k, "%u", i);
            if  (!k) {
                ifBLIN_QX0("no mem");
                errno = ENOMEM;
                ex = -EX_OSERR;
                goto out;
            }
            exx = json2pg_dump(options, val, nnmm, (u_char*)k, ex, ++exx);
        }
        ex = exx;
        break;
    case json_type_string:
        if  (options->flags & JSON2PG_COPY) {
            u_char *m, *n = (u_char*)strdup(json_object_get_string(jobj));

            m = n;
            if  ((e = getCGIparmpass3(options->flags & BLIN_VERMASK, (u_char*)"copy", &n))) {
                ifBLIN_QW0("getCGIparmpass3");
                ex = (ex > 0) ? -e : e;
                goto out;
            }
            if  (options->flags & JSON2PG_COPZ) {
                printf("%u	%u	%s	%s\n", exx, jt, IFNMM, n);
            } else {
                printf("%u	%u	%u	%s	%s\n", exx, ex, jt, nnm, n);
            }
            if  (m != n) free(n);
            free(m);
        } else {
            printf("%s %s=%s~\n", jtnm[jt], IFNM, json_object_get_string(jobj));
        }
        ex = exx;
        break;
    default:
        if  (options->flags & JSON2PG_COPY) {
            u_char *m, *n = (u_char*)strdup(json_object_get_string(jobj));

            m = n;
            if  ((e = getCGIparmpass3(options->flags & BLIN_VERMASK, (u_char*)"copy", &n))) {
                ifBLIN_QW0("getCGIparmpass3");
                ex = (ex > 0) ? -e : e;
                goto out;
            }
            if  (options->flags & JSON2PG_COPZ) {
                printf("%u	%u	%s	%s\n", exx, jt, IFNMM, n);
            } else {
                printf("%u	%u	%u	%s	%s\n", exx, ex, jt, nnm, n);
            }
            if  (m != n) free(n);
            free(m);
        } else {
            printf("%2d=unkn %s.%s=%s~\n", jt, IFNM, nnm, json_object_get_string(jobj));
        }
        ex = exx;
        break;
    }
out:
    if  (pnm && (pnm != nnm)) free(nnm);
    ifBLIN_QX2( "- %u\n", ex);
    return(ex);
}

static void
usage(int ex) {
    fprintf( stderr
           , "json2pg " VERS "  " DATE "\n"
             "    -c[c]     - COPY object recursive/plain\n"
             "    -D[D[D]]  - dump object\n"
             "    -q        - quiet\n"
             "    -v        - verbose\n"
           )
    ;
    exit(ex);
}

static u_char *
rediropt(json2pg_cf *options, babolo_opt **opt, int redirenv) {
    ifBLIN_FLAG = options->flags;
    u_char *o;
    char *e;

    if  (!(o = (u_char*)babolo_getoptarg(opt, 0))) {
        ifBLIN_QW0("No argument");
        usage(EX_USAGE);
    } else if ((redirenv > 0) && (!(o = (u_char*)getenv(e = (char*)o)))) {
        ifBLIN_QX0("No ENV %s", e);
        usage(EX_USAGE);
    }
    return(o);
}

int
main(int argc, char *argv[], char **envp) {
#   define blin_internal_flags (options->flags)
    int c, redirenv, ex = EX_OK;
    struct json_tokener *tok;
    mife_descriptor *mdi;
    json2pg_cf *options;
    BLIN_flag subflags;
    json_object *jobj;
    babolo_parm *bp;
    babolo_opt *opt;
    char *flar, *o;
    ssize_t jsize;
    u_char *oo;

    if  (!(options = calloc(1, sizeof(*options)))) {
        warnx("Malloc failed #1");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    options->flags = BLIN_VER1;
    if  (babolo_openopt(Bpars_NONU, &opt, argc, argv, flar = "cDEqvz:Z:?")) {
        ifBLIN_QW0("openopt");
        ex = EX_SOFTWARE;
        goto out;
    }
    for (redirenv = 0; (c = babolo_getopt(&opt)) != 0; redirenv--) {
        if  (redirenv < 0) redirenv = 0;
        switch (c) {
        case 'c': options->flags |= (options->flags & JSON2PG_COPY) << 1;
                ; options->flags |= JSON2PG_COPY;
                ; break;
        case 'D': options->flags |= (options->flags & (JSON2PG_DUMP | JSON2PG_ODMP)) << 1;
                ; options->flags |= JSON2PG_DUMP;
                ; break;
        case 'E': redirenv = 2;                                                                  break;
        case 'q': options->flags &= ~BLIN_VERMASK;                                               break;
        case 'v': BLIN_VERBOSE(options->flags);                                                  break;
        case 'z': oo = rediropt(options, &opt, redirenv);
                ; bp = babolo_getparm(Bpars_NOEN | Bpars_NOAL | Bpars_NONU, &oo, (u_char*)"", 0);
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 'Z': oo = rediropt(options, &opt, redirenv);
                ; bp = babolo_getparm(Bpars_NOAL | Bpars_NONU, &oo, (u_char*)"", 0);
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c", c);
                ; usage(EX_USAGE);
    }   }
    subflags = (((BLIN_VER1 | BLIN_VER2 ) & options->flags) | (options->flags >> 1)) & BLIN_VERMASK;
    if  (!(mdi = mife_opef(subflags, fileno(stdin)))) {
        ifBLIN_QW0("mife_opef stdin");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > (jsize = mife_read(mdi, 0, 0))) {
        ifBLIN_QW0("mife_read stdin");
        ex = EX_IOERR;
        goto out;
    }
    if  (!(o = mife_get(mdi, 0))) {
        ifBLIN_QW0("mife_get stdin");
        ex = EX_IOERR;
        goto out;
    }
    if  (!(tok = json_tokener_new())) {
        ifBLIN_QW0("json_tokener_new");
        errno = EDOOFUS;
        ex = EX_SOFTWARE;
        goto out;
    }
    jobj = json_tokener_parse_ex(tok, o, jsize);
    if  (json_tokener_success != tok->err) {
        ifBLIN_QX0("json_tokener_parse_ex: %d=%s", tok->err, json_tokener_errors[tok->err]);
        errno = EDOOFUS;
        ex = EX_SOFTWARE;
        goto out;
    } else if (!jobj) {
        ifBLIN_QX0("json_tokener_parse_ex: unknown error");
        errno = EDOOFUS;
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (tok->char_offset < jsize) {
        ifBLIN_QX1("Extra symbols=%s~", o + tok->char_offset);
    }
    if  (options->flags & (JSON2PG_DUMP | JSON2PG_COPY)) {
        if  (0 > (ex = json2pg_dump(options, jobj, NULL, NULL, 0, 1))) {
            ifBLIN_QX0("json2pg_dump");
            ex = -ex;
            goto out;
        } else {
            ex = 0;
    }   }
out:
    exit(ex);
}
